{ ******************************************************************
  Monte-Carlo simulation of the statistical distribution
  of the regression parameters for the exponential model:
  Y = B(1) * Exp(- B(2) * X)
  ****************************************************************** }

program mcsim;

uses
  utypes, uinterv, umcmc, unlfit, ustrings, uplot;

{$define Vector array[Lb..Ub : Integer] of Float}

const
  FuncName = 'Y = B(1) * Exp(- B(2) * X)';

const
  NCycles  =   10;         { Number of cycles }
  MaxSim   = 1000;         { Max nb of simulations at each cycle }
  SavedSim = 1000;         { Nb of simulations to be saved }
  MCFile   = 'mcsim.txt';  { File for storing simulation results }

const
  N = 10;  { Number of points }

{ Data }
var
  X : array[1..N] of Float = (  1,   2,   3,   4,   5,   6,  7,  8,  9, 10);
  Y : array[1..N] of Float = (416, 319, 244, 188, 144, 113, 85, 66, 50, 41);

{ Parameter vector }
var B : array[1..2] of Float;

{ Variance-covariance matrix }
var V : array[1..2, 1..2] of Float;

{ Simulated parameters }
var B1, B2 : array[1..SavedSim] of Float;

{ ******************************************************************
  Subprograms
  ****************************************************************** }

function RegFunc(X : Float; var B : Vector) : Float;
begin
  RegFunc := B[1] * Exp(- B[2] * X);
end;

procedure WriteResults;
{ ------------------------------------------------------------------
  Writes results to screen
  ------------------------------------------------------------------ }

var
  Line1,
  Line2 : String;   { Separating lines }
  SB    : Float;    { Standard deviations of parameters }
  I     : Integer;  { Loop variable }

begin
  Line1 := StrChar(73, '-');
  Line2 := StrChar(73, '=');

  WriteLn(Line2);
  WriteLn('Monte-Carlo simulation : ', FuncName);
  WriteLn(Line1);

  WriteLn('Parameter    Est.value         Std.dev.        ');

  WriteLn(Line1);

  for I := 1 to 2 do
    begin
      SB := Sqrt(V[I,I]);
      WriteLn('B(', I:1, ')', B[I]:17:8, SB:17:8);
    end;

  WriteLn(Line2);
end;

procedure PlotGraph;
{ ------------------------------------------------------------------
  Plots simulation results
  ------------------------------------------------------------------ }

var
  Xmin, Xmax, Xstep : Float;    { Ox scale }
  Ymin, Ymax, Ystep : Float;    { Oy scale }

begin
  if not InitGraphics(9, 2, '../../chr') then
    begin
      Writeln('Unable to set graphic mode');
      Exit;
    end;

  SetWindow(15, 85, 15, 85, True);

  AutoScale(B1, LinScale, Xmin, Xmax, Xstep);
  AutoScale(B2, LinScale, Ymin, Ymax, Ystep);

  SetOxScale(LinScale, Xmin, Xmax, Xstep);
  SetOyScale(LinScale, Ymin, Ymax, Ystep);

  SetGraphTitle('Monte-Carlo simulation : ' + FuncName);
  SetOxTitle('B(1)');
  SetOyTitle('B(2)');

  SetFormat(10, 4, False, False);

  PlotOxAxis;
  PlotOyAxis;

  WriteGraphTitle;

  SetClipping(True);

  SetLineParam(1, 0, 0, 0);  { Don't connect points }
  PlotCurve(B1, B2, 1);

  Readln;

  LeaveGraphics;
end;

var
  F    : Text;     { Output file }
  Iter : Integer;  { Iteration number }
  I    : Integer;  { Loop variable }

begin
  { Initialize parameters }
  SetParamBounds(1, 100, 1000);
  SetParamBounds(2, 0.1, 1);

  { Set Metropolis-Hastings parameters }
  InitMHParams(NCycles, MaxSim, SavedSim);

  { Set output file and numeric format }
  SetMCFile(MCFile);
  SetFormat(10, 4, False, True);

  { Perform simulation }
  SimFit(RegFunc, X, Y, B, V);

  { Retrieve simulation results into vectors B1 and B2 }
  Assign(F, MCFile);
  Reset(F);
  for I := 1 to SavedSim do
    ReadLn(F, Iter, B1[I], B2[I]);
  Close(F);

  { Write results }
  WriteResults;

  { Plot curve }
  PlotGraph;
end.
