{ ******************************************************************
  This program simulates a multinormal distribution by Markov Chain
  Monte Carlo (MCMC) using the Hastings-Metropolis algorithm.

  Although MCMC is best used when there is no direct way to simulate
  the distribution, it is used here for demonstration purposes since
  its results can be compared to those of the direct method (program
  ranmul.pas).

  The pdf P(X) of the multinormal distribution is such that:

                        P(X) = C * Exp(- F(X) / T)

  where F(X) = (X - M)' * V^(-1) * (X - M)

        C = 1/sqrt(|V| * (2*Pi)^N)

        T = 2

        M is the mean vector and V the variance-covariance matrix of
        the distribution. N is the dimension of the distribution.

        The constant C is not used in the simulation.

  The results are stored in the output file testmcmc.txt
  ****************************************************************** }

program testmcmc;

uses
  utypes, ugausjor, umcmc;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define Matrix array[Lb1..Ub1 : Integer; Lb2..Ub2 : Integer] of Float}

const
  N        = 3;     { Size of distribution }
  Temp     = 2.0;   { Temperature }
  NCycles  = 10;    { Number of cycles }
  MaxSim   = 1000;  { Max nb of simulations }
  SavedSim = 1000;  { Nb of saved simulations }

var  
  M     : array[1..N] of Float = (1.0, 2.0, 3.0);  { Means }
  S     : array[1..N] of Float = (0.1, 0.2, 0.3);  { S.D. }

  R     : array[1..N, 1..N] of Float = 
          ((1.00,  0.25,  0.50),
           (0.25,  1.00, -0.25),
           (0.50, -0.25,  1.00));      { Correlation matrix }
  
  Msim  : array[1..N] of Float;        { Simulated means }
  X     : array[1..N] of Float;        { Random vector }
  X_min : array[1..N] of Float;        { Coordinates of minimum }

  V     : array[1..N, 1..N] of Float;  { Original var. cov. matrix }
  V_inv : array[1..N, 1..N] of Float;  { Inverse var. cov. matrix }
  Vsim  : array[1..N, 1..N] of Float;  { Simulated var. cov. matrix }
  
  Xmat  : array[1..SavedSim, 1..N] of Float;  { Simulated vectors }

  F     : Text;     { Output file }
  F_min : Float;    { Value of F(X) at minimum }
  I, J  : Integer;  { Loop variables }
  Det   : Float;    { Determinant of var-cov. matrix }

function ObjFunc(var X : Vector) : Float;
{ Computes the function F(X) }
var
  Sum1, 
  Sum2 : Float;
  I, J : Integer;
  D    : array[Lb..Ub] of Float;
begin
  for I := 1 to N do
    D[I] := X[I] - M[I];

  Sum1 := 0.0;
  for I := 1 to N do
    Sum1 := Sum1 + V_inv[I,I] * Sqr(D[I]);

  Sum2 := 0.0;
  for I := 2 to N do
    for J := 1 to Pred(I) do
      Sum2 := Sum2 + V_inv[I,J] * D[I] * D[J];

  ObjFunc := Sum1 + 2.0 * Sum2;
end;

procedure WriteResults(    Title : String; 
                       var M     : Vector;
                       var V     : Matrix);
var
  I, J : Integer;
  S    : array[Lb..Ub] of Float;
  R    : Float;
begin
  WriteLn;
  WriteLn(Title);
  WriteLn;

  WriteLn('      Mean      S.D.');
  WriteLn('--------------------');

  for I := 1 to N do
    begin
      S[I] := Sqrt(V[I,I]);
      Writeln(M[I]:10:4, S[I]:10:4);
    end;

  WriteLn;
  WriteLn('Correlation matrix:');
  WriteLn;

  for I := 2 to N do
    begin
      for J := 1 to Pred(I) do
        begin
          R := V[I,J] / (S[I] * S[J]);
          Write(R:10:4);
        end;
      WriteLn;
    end;
end;

procedure WriteOutputFile(Title : String; var Xmat : Matrix);
var
  F    : Text;
  I, J : Integer;
begin
  Assign(F, 'testmcmc.txt');
  Rewrite(F);

  WriteLn(F, Title);
  Write(F, ' Iter');
  for I := 1 to N do
    Write(F, '        X', I);
  WriteLn(F);

  for I := 1 to SavedSim do
    begin
      Write(F, I:5);
      for J := 1 to N do
        Write(F, Xmat[I,J]:10:4);
      WriteLn(F);
    end;

  Close(F);
end;

begin
  for I := 1 to N do
    V[I,I] := Sqr(S[I]);  { Variances }
  
  for I := 2 to N do
    for J := 1 to Pred(I) do
      begin
        V[I,J] := R[I,J] * S[I] * S[J];  { Covariances }
        V[J,I] := V[I,J];
      end;

  { Initialize inverse var-cov. matrix }
  for I := 1 to N do
    for J := 1 to N do
      V_inv[I,J] := V[I,J];

  { Compute the inverse of the variance-covariance matrix }
  GaussJordan(V_inv, Det);

  if MathErr = MatSing then
    begin
      WriteLn('Variance-covariance matrix is singular!');
      Exit;
    end;

  { Initialize Metropolis-Hastings parameters }
  InitMHParams(NCycles, MaxSim, SavedSim);

  { Initialize the mean vector and the variance-covariance matrix.
    For the sake of demonstration we start at a distance from the
    true mean and with enhanced standard deviations. }
  for I := 1 to N do
    begin
      Msim[I] := 3.0 * M[I];
      Vsim[I,I] := 10.0 * V[I,I];
    end;

  { Perform Metropolis-Hastings simulations }
  Write('Running. Please wait...');

  Hastings(ObjFunc, Temp, Msim, Vsim, Xmat, X_min, F_min);

  if MathErr = MatOk then
    begin
      WriteResults('Original distribution', M, V);
      WriteResults('Simulated distribution', Msim, Vsim);
      WriteOutputFile('Simulated distribution', Xmat); 
    end
  else
    WriteLn('Variance-covariance matrix is not positive definite!');
end.

