{ ******************************************************************
  This program performs a nonlinear least squares fit of a
  gamma distribution function :

     y = a * (x - b)^c * exp[ - (x - b) / d ]

  The parameters are: B(1) = a, B(2) = b, B(3) = c, B(4) = d

  The name of the input file (default extension = .DAT) is passed
  on the command line. The structure of this file is described in
  REGUTILS.PAS

  The file GAMMA.DAT is an example provided by Chris Rorden
  ****************************************************************** }

program regamma;

uses
  utypes, uinterv, unlfit, ugamfit,
  uregtest, uinvbeta, uplot, regutils;

const
  MaxIter = 1000;    { Max. number of iterations }
  Tol     = 1.0E-4;  { Required precision }
  Alpha   = 0.05;    { Significance level }

var
  InFName : String;    { Name of input file }
  Title   : String;    { Title of study }
  XName,
  YName   : String;    { Names of variable }
  N       : Integer;   { Number of points }
  Test    : TRegTest;  { Regression tests }
  Sr      : Float;     { Residual standard dev. }
  Tc      : Float;     { Critical t value }
  Fc      : Float;     { Critical F value }
  I       : Integer;   { Loop variable }

  B : array[1..4] of Float;        { Regression parameters }
  V : array[1..4, 1..4] of Float;  { Variance-covariance matrix }

procedure ReadCmdLine(var InFName : String);
{ ------------------------------------------------------------------
  Reads command line
  ------------------------------------------------------------------ }
  begin
    { Name of input file }
    InFName := ParamStr(1);
    if InFName = '' then
      begin
        InFName := 'gamma.dat';
        Exit;
      end;
    if Pos('.', InFName) = 0 then InFName := InFName + '.dat';
  end;

procedure SetBounds;
{ ------------------------------------------------------------------
  Initialize parameters and bounds
  ------------------------------------------------------------------ }
  var
    I : Integer;
  begin
    for I := 1 to 4 do
      begin
        SetParamBounds(I, 0, 1000);  { All parameters must be > 0 }
        B[I] := 0.0;
      end;
  end;

{ *************************** Main program ************************* }

begin
  { Read command line  }
  ReadCmdLine(InFName);

  { Read number of points }
  N := ReadNPoints(InFName, 5);
  if N < 0 then Halt;

  { Dimension arrays }
  var X, Y, S, Ycalc : array[1..N] of Float;

  { Read input file }
  if ReadInputFile(InFName, Title, XName, YName, X, Y) <> 0 then
    begin
      WriteLn('Error reading file ', InFName);
      Halt;
    end;

  { Set parameter bounds }
  SetBounds;

  { Perform regression }
  GammaFit(X, Y, MaxIter, Tol, B, V);

  if MathErr <> MatOk then
    begin
      Writeln('Unable to fit curve! (MathErr = ', MathErr, ')');
      Halt;
    end;

  { Compute predicted Y values }
  for I := 1 to N do
    Ycalc[I] := GammaFit_Func(X[I], B);

  { Update var-cov. matrix and compute statistical tests }
  RegTest(Y, Ycalc, V, Test);

  { Compute Student's t and Snedecor's F }
  Tc := InvStudent(Test.Nu2, 1 - 0.5 * Alpha);
  Fc := InvSnedecor(Test.Nu1, Test.Nu2, 1 - Alpha);

  { Update standard deviations }
  Sr := Sqrt(Test.Vr);
  for I := 1 to N do
    S[I] := Sr;

  { Write results }
  WriteResults(Title, N, X, Y, S, Ycalc, B, V, Test, Alpha, Tc, Fc);

  { Plot curve }
  PlotGraph(GammaFit_Func, Title, X, Y, B);
end.
