{ ******************************************************************
  Optimization by Genetic Algorithm
  ******************************************************************
  Ref.:  E. Perrin, A. Mandrille, M. Oumoun, C. Fonteix & I. Marc
         Optimisation globale par strategie d'evolution
         Technique utilisant la genetique des individus diploides
         Recherche operationnelle / Operations Research
         1997, 31, 161-201
  Thanks to Magali Camut for her contribution
  ****************************************************************** }

unit ugenalg;

interface

uses
  utypes, uswap, urandom;

{$define Vector array[Lb..Ub : Integer] of Float}

procedure InitGAParams(NP, NG : Integer; SR, MR, HR : Float);
{ ------------------------------------------------------------------
  Initialize Genetic Algorithm parameters
  ------------------------------------------------------------------
    NP : Population size
    NG : Max number of generations
    SR : Survival rate
    MR : Mutation rate
    HR : Proportion of homozygotes
  ------------------------------------------------------------------ }

procedure GA_CreateLogFile(LogFileName : String);
{ ------------------------------------------------------------------
  Initialize log file
  ------------------------------------------------------------------ }

procedure GenAlg(    Func          : TFuncNVar;
                 var X, Xmin, Xmax : Vector;
                 var F_min         : Float);
{ ------------------------------------------------------------------
  Minimization of a function of several variables
  by genetic algorithm
  ------------------------------------------------------------------
  Input parameters : Func = objective function to be minimized
                     X    = initial minimum coordinates
                     Xmin = minimum value of X
                     Xmax = maximum value of X
  ------------------------------------------------------------------
  Output parameters: X     = refined minimum coordinates
                     F_min = function value at minimum
  ------------------------------------------------------------------ }

implementation

var
  GA_NP : Integer = 200;  { Population size }
  GA_NG : Integer =  40;  { Max number of generations }
  GA_SR : Float   = 0.6;  { Survival rate }
  GA_MR : Float   = 0.1;  { Mutation rate }
  GA_HR : Float   = 0.5;  { Proportion of homozygotes }

  WriteLogFile : Boolean = False;

var
  LogFile : Text;

procedure InitGAParams(NP, NG : Integer; SR, MR, HR : Float);
begin
  if NP > 0 then GA_NP := NP;
  if NG > 0 then GA_NG := NG;

  if (SR > 0.0) and (SR < 1.0) then GA_SR := SR;
  if (MR > 0.0) and (MR < 1.0) then GA_MR := MR;
  if (HR > 0.0) and (HR < 1.0) then GA_HR := HR;
end;

procedure GA_CreateLogFile(LogFileName : String);
begin
  Assign(LogFile, LogFileName);
  Rewrite(LogFile);
  Writeln(LogFile, 'Genetic Algorithm');
  Writeln(LogFile, ' Iter          F ');
  WriteLogFile := True;
end;

procedure GenAlg(    Func          : TFuncNVar;
                 var X, Xmin, Xmax : Vector;
                 var F_min         : Float);
var
  I, NS, Iter  : Integer;
  C1, C2, D, P : array[1 .. GA_NP, Lb .. Ub] of Float;
  F            : array[1 .. GA_NP] of Float;
  Range        : array[Lb .. Ub] of Float;

  procedure Mutate(I : Integer);
  { Mutate individual I }
  var
    J : Integer;
  begin
    for J := Lb to Ub do
      begin
        C1[I,J] := Xmin[J] + RanGen3 * Range[J];
        C2[I,J] := Xmin[J] + RanGen3 * Range[J];
        D[I,J] := RanGen3;
        P[I,J] := D[I,J] * C1[I,J] + (1.0 - D[I,J]) * C2[I,J];
      end;
  end;

  procedure Cross(I1, I2, I : Integer);
  { Cross two individuals I1 and I2 --> new individual I }
  var
    J, K : Integer;
  begin
    for J := Lb to Ub do
      begin
        if RanGen3 < 0.5 then K := I1 else K := I2;
        C1[I,J] := C1[K,J];

        if RanGen3 < 0.5 then K := I1 else K := I2;
        C2[I,J] := C2[K,J];

        D[I,J] := RanGen3;

        P[I,J] := D[I,J] * C1[I,J] + (1.0 - D[I,J]) * C2[I,J];
      end;
  end;

  procedure Homozygote(I : Integer);
  { Make individual I homozygous }
  var
    J : Integer;
  begin
    for J := Lb to Ub do
      begin
        C1[I,J] := P[I,J];
        C2[I,J] := P[I,J];
      end;
  end;

  function GA_Func(I : Integer) : Float;
  { Computes objective function for individual I }
  var
    J : Integer;
    X : array[Lb..Ub] of Float;
  begin
    for J := Lb to Ub do
      X[J] := P[I,J];

    GA_Func := Func(X);
  end;

  procedure CompFunc;
  { Computes function values }
  var
    I, J, K : Integer;
    A       : Float;
  begin
    for I := 1 to GA_NP do
      F[I] := GA_Func(I);

    { Sort population according to function values }
    for I := 1 to GA_NP - 1 do
      begin
        K := I;
        A := F[I];

        for J := I + 1 to GA_NP do
          if F[J] < A then
            begin
              K := J;
              A := F[J];
            end;

        FSwap(F[I], F[K]);

        for J := Lb to Ub do
          begin
            FSwap(C1[I,J], C1[K,J]);
            FSwap(C2[I,J], C2[K,J]);
            FSwap(D[I,J], D[K,J]);
            FSwap(P[I,J], P[K,J]);
          end;
      end;

    { Update log file if necessary }
    if WriteLogFile then
      Writeln(LogFile, Iter:5, F[1]:12);

    { Update minimum }
    if F[1] < F_min then
      begin
        F_min := F[1];
        for J := Lb to Ub do
          X[J] := P[1,J];
      end;

    Inc(Iter);
  end;

  procedure GenPop;
  { Generates new population }
  var
    I, I1, I2 : Integer;
    F0        : Float;
  begin
    for I := NS + 1 to GA_NP do
      begin
        I1 := Trunc(RanGen3 * NS) + 1;

        repeat
          I2 := Trunc(RanGen3 * NS) + 1
        until I2 <> I1;

        F0 := Max(F[I1], F[I2]);

        repeat
          Cross(I1, I2, I);
        until GA_Func(I) <= F0;
      end;

    for I := 1 to GA_NP do
      begin
        if RanGen3 < GA_MR then
          Mutate(I);
        if RanGen3 < GA_HR then
          Homozygote(I);
      end;
  end;

begin
  SetErrCode(OptOk);

  { Initialize the random number generator
    using the standard generator }
  Randomize;
  InitMT(Trunc(Random * 1.0E+8));

  for I := Lb to Ub do
    Range[I] := Xmax[I] - Xmin[I];

  NS := Trunc(GA_NP * GA_SR);  { Number of survivors }

  Iter := 0;
  F_min := MaxNum;

  for I := 1 to GA_NP do
    Mutate(I);

  CompFunc;

  for I := 1 to GA_NG do
    begin
      GenPop;
      CompFunc;
    end;

  if WriteLogFile then
    begin
      Close(LogFile);
      WriteLogFile := False;
    end;
end;

end.
