{ ******************************************************************
  This unit fits the Hill equation :

                                  Ymax . x^n
                              y = ----------
                                  K^n + x^n

  ****************************************************************** }

unit uhillfit;

interface

uses
  utypes, umath, umeansd, ulinfit, unlfit;

{$i regdef.inc}

procedure HillFit(var X, Y : Vector;
                  MaxIter  : Integer;
                  Tol      : Float;
                  var B    : BVector;
                  var V    : VMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of model
  ------------------------------------------------------------------
  Input parameters:  X, Y    = point coordinates
                     MaxIter = max. number of iterations
                     Tol     = tolerance on parameters
  Output parameters: B       = regression parameters
                     V       = inverse matrix
  ------------------------------------------------------------------ }

procedure WHillFit(var X, Y, S : Vector;
                   MaxIter     : Integer;
                   Tol         : Float;
                   var B       : BVector;
                   var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted fit of model
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function HillFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Returns the value of the regression function at point X
  ------------------------------------------------------------------ }

implementation

function HillFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Computes the regression function at point X
  B is the vector of parameters, such that :

  B[0] = Ymax     B[1] = K     B[2] = n
  ------------------------------------------------------------------ }
begin
  if X = 0.0 then
    if B[2] > 0.0 then HillFit_Func := 0.0 else HillFit_Func := B[0]
  else
    { Compute function according to y = Ymax / [1 + (K/x)^n] }
    HillFit_Func := B[0] / (1.0 + (B[1] / X) ** B[2]);
end;

procedure HillFit_Deriv(X, Y : Float; var B, D : BVector);
{ ------------------------------------------------------------------
  Computes the derivatives of the regression function at point (X,Y)
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter
  ------------------------------------------------------------------ }
var
  Q, R, S : Float;
begin
  if X = 0.0 then
    begin
      if B[2] > 0.0 then D[0] := 0.0 else D[0] := 1.0;
      D[1] := 0.0;
      D[2] := 0.0;
    end
  else
    begin
      Q := (B[1] / X) ** B[2];  { (K/x)^n }
      R := 1.0 / (1.0 + Q);     { 1 / [1 + (K/x)^n] }
      S := - Y * R * Q;         { -Ymax.(K/x)^n / [1 + (K/x)^n]^2 }

      { dy/dYmax = 1 / [1 + (K/x)^n] }
      D[0] := R;

      { dy/dK = -Ymax.(K/x)^n.(n/K)/[1 + (K/x)^n]^2 }
      D[1] := S * B[2] / B[1];

      { dy/dn = -Ymax.(K/x)^n.Ln(K/x)/[1 + (K/x)^n]^2 }
      D[2] := S * Log(B[1] / X);
    end;
end;

procedure ApproxFit(Mode : TRegMode; var X, Y, S : Vector;
                     var B : BVector);
{ ------------------------------------------------------------------
  Approximate fit of the Hill equation by linear regression:
  Ln(Ymax/y - 1) = n.Ln(K) - n.Ln(x)
  ------------------------------------------------------------------
  Input :  Mode = OLS for unweighted regression, WLS for weighted
           X, Y = point coordinates
           S    = standard deviations of Y values
  Output : B    = estimated regression parameters
  ------------------------------------------------------------------ }
var
  Ymax : Float;    { Estimated value of Ymax }
  P    : Integer;  { Number of points for linear regression }
  K    : Integer;  { Loop variable }

  A : array[0..1] of Float;        { Linear regression param. }
  V : array[0..1, 0..1] of Float;  { Variance-covariance matrix }

begin
  Ymax := 1.1 * MaxVal(Y);

  P := 0;
  for K := Lb to Ub do
    if (X[K] > 0.0) and (Y[K] > 0.0) then
      Inc(P);

  if P < 3 then
    begin
      SetErrCode(MatSing);
      Exit;
    end;

  var X1, Y1, S1 : array[1..P] of Float;

  P := 0;
  for K := Lb to Ub do
    if (X[K] > 0.0) and (Y[K] > 0.0) then
      begin
        Inc(P);
        X1[P] := Log(X[K]);
        Y1[P] := Log(Ymax / Y[K] - 1.0);
        S1[P] := 1.0 / (Y[K] * (1.0 - Y[K] / Ymax));
        if Mode = WLS then S1[P] := S1[P] * S[K];
      end;

  WLinFit(X1, Y1, S1, A, V);

  if MathErr = MatOk then
    begin
      B[0] := Ymax;
      B[1] := Expo(- A[0] / A[1]);
      B[2] := - A[1];
    end;
end;

procedure GenHillFit(Mode        : TRegMode;
                     var X, Y, S : Vector;
                     MaxIter     : Integer;
                     Tol         : Float;
                     var B       : BVector;
                     var V       : VMatrix);
begin
  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    ApproxFit(Mode, X, Y, S, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(HillFit_Func, HillFit_Deriv, X, Y,
                MaxIter, Tol, B, V);
    WLS : WNLFit(HillFit_Func, HillFit_Deriv, X, Y, S,
                 MaxIter, Tol, B, V);
  end;
end;

procedure HillFit(var X, Y : Vector;
                  MaxIter  : Integer;
                  Tol      : Float;
                  var B    : BVector;
                  var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenHillFit(OLS, X, Y, S, MaxIter, Tol, B, V);
end;

procedure WHillFit(var X, Y, S : Vector;
                   MaxIter     : Integer;
                   Tol         : Float;
                   var B       : BVector;
                   var V       : VMatrix);
begin
  GenHillFit(WLS, X, Y, S, MaxIter, Tol, B, V);
end;

end.
