{ ******************************************************************
  This unit fits the increasing exponential :

                   y = Ymin + A.[1 - exp(-k.x)]

  ****************************************************************** }

unit uiexpfit;

interface

uses
  utypes, umath, umeansd, ulinfit, unlfit;

{$i regdef.inc}

procedure IncExpFit(var X, Y : Vector;
                    ConsTerm : Boolean;
                    MaxIter  : Integer;
                    Tol      : Float;
                    var B    : BVector;
                    var V    : VMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of model
  ------------------------------------------------------------------
  Input parameters:  X, Y     = point coordinates
                     ConsTerm = flag for presence of constant term (Ymin)
                     MaxIter  = max. number of iterations
                     Tol      = tolerance on parameters
  Output parameters: B        = regression parameters
                     V        = inverse matrix
  ------------------------------------------------------------------ }

procedure WIncExpFit(var X, Y, S : Vector;
                     ConsTerm    : Boolean;
                     MaxIter     : Integer;
                     Tol         : Float;
                     var B       : BVector;
                     var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted fit of model
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function IncExpFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Returns the value of the regression function at point X
  ------------------------------------------------------------------ }

implementation

var
  gConsTerm : Boolean;

function IncExpFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Computes the regression function at point X
  B is the vector of parameters, such that :

  B[0] = Ymin     B[1] = A     B[2] = k
  ------------------------------------------------------------------ }
begin
  if gConsTerm then
    IncExpFit_Func := B[0] + B[1] * (1.0 - Expo(- B[2] * X))
  else
    IncExpFit_Func := B[1] * (1.0 - Expo(- B[2] * X));
end;

procedure IncExpFit_Deriv(X, Y : Float; var B, D : BVector);
{ --------------------------------------------------------------------
  Computes the derivatives of the regression function at point (X, Y)
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter.
  -------------------------------------------------------------------- }
var
  E : Float;
begin
  if gConsTerm then
    D[0] := 1.0;          { dy/dYmin = 1 }
  E := Expo(- B[2] * X);  { exp(-k.x) }
  D[1] := 1.0 - E;        { dy/dA = 1 - exp(-k.x) }
  D[2] := B[1] * X * E;   { dy/dk = A.x.exp(-k.x) }
end;

procedure ApproxFit(Mode : TRegMode; var X, Y, S : Vector;
                     var B : BVector);
{ ------------------------------------------------------------------
  Approximate fit of the increasing exponential by linear regression
  Ln(Ymax - Y) = Ln(A) - k.X with A = Ymax - Ymin
  ------------------------------------------------------------------
  Input :  Mode = OLS for unweighted regression, WLS for weighted
           X, Y = point coordinates
           S    = weights
  Output : B    = estimated regression parameters
  ------------------------------------------------------------------ }
var
  Ymax   : Float;
  Z      : Float;                       { Ymax - Y }
  K      : Integer;                     { Loop variable }
  Y1, S1 : array[Lb..Ub] of Float;      { Transformed coord. and SD }
  A      : array[0..1] of Float;        { Linear regression param. }
  V      : array[0..1, 0..1] of Float;  { Variance-covariance matrix }

begin
  { Estimation of Ymin }
  if gConsTerm then
    B[0] := 0.9 * MinVal(Y);

  { Estimation of Ymax = Ymin + A }
  Ymax := 1.1 * MaxVal(Y);

  for K := Lb to Ub do
    begin
      Z := Ymax - Y[K];
      Y1[K] := Ln(Z);
      S1[K] := 1.0 / Z;
      if Mode = WLS then S1[K] := S1[K] * S[K];
    end;

  WLinFit(X, Y1, S1, A, V);

  if MathErr = MatOk then
    begin
      B[1] := Exp(A[0]);
      B[2] := - A[1];
    end;
end;

procedure GenIncExpFit(Mode        : TRegMode;
                       var X, Y, S : Vector;
                       ConsTerm    : Boolean;
                       MaxIter     : Integer;
                       Tol         : Float;
                       var B       : BVector;
                       var V       : VMatrix);
begin
  gConsTerm := ConsTerm;

  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    ApproxFit(Mode, X, Y, S, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(IncExpFit_Func, IncExpFit_Deriv, X, Y,
                       MaxIter, Tol, B, V);
    WLS : WNLFit(IncExpFit_Func, IncExpFit_Deriv, X, Y, S,
                        MaxIter, Tol, B, V);
  end;
end;

procedure IncExpFit(var X, Y : Vector;
                    ConsTerm : Boolean;
                    MaxIter  : Integer;
                    Tol      : Float;
                    var B    : BVector;
                    var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenIncExpFit(OLS, X, Y, S, ConsTerm, MaxIter, Tol, B, V);
end;

procedure WIncExpFit(var X, Y, S : Vector;
                     ConsTerm    : Boolean;
                     MaxIter     : Integer;
                     Tol         : Float;
                     var B       : BVector;
                     var V       : VMatrix);
begin
  GenIncExpFit(WLS, X, Y, S, ConsTerm, MaxIter, Tol, B, V);
end;

end.
