{ ******************************************************************
  Khi-2 and Woolf tests
  ****************************************************************** }

unit ukhi2;

interface

uses
  utypes;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define IntVector array[LbI..UbI : Integer] of Integer}

{$define IntMatrix array[LbI1..UbI1 : Integer; LbI2..UbI2 : Integer] of Integer}

procedure Khi2_Conform(var Obs     : IntVector;
                       var Calc    : Vector;
                           N_estim : Integer;
                       var Khi2    : Float;
                       var DoF     : Integer);
{ ------------------------------------------------------------------
  Khi-2 test for conformity
  ------------------------------------------------------------------ }

procedure Khi2_Indep(var Obs  : IntMatrix;
                     var Khi2 : Float;
                     var DoF  : Integer);
{ ------------------------------------------------------------------
  Khi-2 test for independence
  ------------------------------------------------------------------ }

procedure Woolf_Conform(var Obs     : IntVector;
                        var Calc    : Vector;
                            N_estim : Integer;
                        var G       : Float;
                        var DoF     : Integer);
{ ------------------------------------------------------------------
  Woolf test for conformity
  ------------------------------------------------------------------ }

procedure Woolf_Indep(var Obs : IntMatrix;
                      var G   : Float;
                      var DoF : Integer);
{ ------------------------------------------------------------------
  Woolf test for independence
  ------------------------------------------------------------------ }

implementation

type
  TTest = (Pearson, Woolf);

procedure Test_Conform(    Test    : TTest;
                       var Obs     : IntVector;
                       var Calc    : Vector;
                           N_estim : Integer;
                       var Result  : Float;
                       var DoF     : Integer);

var
  I, N_cls : Integer;

begin
  if (Lb <> 1) or (LbI <> 1) or (Ub <> UbI) then
    begin
      SetErrCode(MatErrDim);
      Exit;
    end;

  SetErrCode(FOk);

  N_cls := Ub;
  Result := 0.0;

  case Test of
    Pearson : for I := 1 to N_cls do
                Result := Result + Sqr(Obs[I] - Calc[I]) / Calc[I];
    Woolf   : for I := 1 to N_cls do
                Result := Result + Obs[I] * Ln(Obs[I] / Calc[I]);
  end;

  if Test = Woolf then Result := 2.0 * Result;
  DoF := N_cls - N_estim - 1;
end;

procedure Test_Indep(    Test   : TTest;
                     var Obs    : IntMatrix;
                     var Result : Float;
                     var DoF    : Integer);
var
  SumLin       : array[1..UbI1] of Integer;
  SumCol       : array[1..UbI2] of Integer;
  N_lin, N_col : Integer;
  Sum          : Integer;
  Prob, Calc   : Float;
  I, J         : Integer;

begin
  if (LbI1 <> 1) or (LbI2 <> 1) then
    begin
      SetErrCode(MatErrDim);
      Exit;
    end;

  SetErrCode(FOk);

  N_lin := UbI1;
  N_col := UbI2;

  for I := 1 to N_lin do
    begin
      SumLin[I] := 0;
      for J := 1 to N_col do
        SumLin[I] := SumLin[I] + Obs[I,J];
    end;

  for J := 1 to N_col do
    begin
      SumCol[J] := 0;
      for I := 1 to N_lin do
        SumCol[J] := SumCol[J] + Obs[I,J];
    end;

  Sum := 0;
  for I := 1 to N_lin do
    Sum := Sum + SumLin[I];

  Result := 0.0;
  for I := 1 to N_lin do
    begin
      Prob := SumLin[I] / Sum;
      case Test of
        Pearson : for J := 1 to N_col do
                    begin
                      Calc := SumCol[J] * Prob;
                      Result := Result + Sqr(Obs[I,J] - Calc) / Calc;
                    end;
        Woolf   : for J := 1 to N_col do
                    begin
                      Calc := SumCol[J] * Prob;
                      Result := Result + Obs[I,J] * Ln(Obs[I,J] / Calc);
                    end;
      end;
    end;

  if Test = Woolf then Result := 2.0 * Result;
  DoF := Pred(N_lin) * Pred(N_col);
end;

procedure Khi2_Conform(var Obs     : IntVector;
                       var Calc    : Vector;
                           N_estim : Integer;
                       var Khi2    : Float;
                       var DoF     : Integer);
begin
  Test_Conform(Pearson, Obs, Calc, N_estim, Khi2, DoF);
end;

procedure Khi2_Indep(var Obs  : IntMatrix;
                     var Khi2 : Float;
                     var DoF  : Integer);
begin
  Test_Indep(Pearson, Obs, Khi2, DoF);
end;

procedure Woolf_Conform(var Obs     : IntVector;
                        var Calc    : Vector;
                            N_estim : Integer;
                        var G       : Float;
                        var DoF     : Integer);
begin
  Test_Conform(Woolf, Obs, Calc, N_estim, G, DoF);
end;

procedure Woolf_Indep(var Obs : IntMatrix;
                      var G   : Float;
                      var DoF : Integer);
begin
  Test_Indep(Woolf, Obs, G, DoF);
end;

end.
