{ ******************************************************************
  LU decomposition
  ****************************************************************** }

unit ulu;

interface

uses
  utypes, uswap;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define Matrix array[Lb1..Ub1 : Integer; Lb2..Ub2 : Integer] of Float}

procedure LU_Decomp(var A : Matrix);
{ ------------------------------------------------------------------
  LU decomposition. Factors the square matrix A as a product L * U,
  where L is a lower triangular matrix (with unit diagonal terms) and
  U is an upper triangular matrix. This routine is used in conjunction
  with LU_Solve to solve a system of equations.
  ------------------------------------------------------------------
  Input parameters : A = matrix
  ------------------------------------------------------------------
  Output parameter : A = contains the elements of L and U
  ------------------------------------------------------------------
  Possible results : MatOk
                     MatSing
                     MatErrDim
  ------------------------------------------------------------------
  NB : This procedure destroys the original matrix A
  ------------------------------------------------------------------ }

procedure LU_Solve(var A : Matrix; var B, X : Vector);
{ ------------------------------------------------------------------
  Solves a system of equations whose matrix has been transformed by
  LU_Decomp
  ------------------------------------------------------------------
  Input parameters : A = result from LU_Decomp
                     B = constant vector
  ------------------------------------------------------------------
  Output parameter : X = solution vector
  ------------------------------------------------------------------ }

implementation

type
  TIndex(Lb, Ub : Integer) = array[Lb..Ub] of Integer;

var
  Index   : ^TIndex;      { Records the row permutations }
  LbIndex : Integer = 0;  { Lower bound of Index }
  UbIndex : Integer = 0;  { Upper bound of Index }

procedure LU_Decomp(var A : Matrix);
  var
    I, Imax, J, K : Integer;
    Pvt, T, Sum   : Float;
    V             : array[Lb1..Ub1] of Float;
  begin
    if (Lb1 <> Lb2) or (Ub1 <> Ub2) then
      begin
        SetErrCode(MatErrDim);
        Exit;
      end;

    { Reallocate Index if necessary}
    if (Lb1 <> LbIndex) or (Ub1 <> UbIndex) then
      begin
        Dispose(Index);
        LbIndex := Lb1;
        UbIndex := Ub1;
        New(Index, LbIndex, UbIndex);
      end;

    for I := Lb1 to Ub1 do
      begin
        Pvt := 0.0;
        for J := Lb1 to Ub1 do
          if Abs(A[I,J]) > Pvt then
            Pvt := Abs(A[I,J]);
        if Pvt < MachEp then
          begin
            SetErrCode(MatSing);
            Exit;
          end;
        V[I] := 1.0 / Pvt;
      end;

    for J := Lb1 to Ub1 do
      begin
        for I := Lb1 to Pred(J) do
          begin
            Sum := A[I,J];
            for K := Lb1 to Pred(I) do
              Sum := Sum - A[I,K] * A[K,J];
            A[I,J] := Sum;
          end;
        Imax := 0;
        Pvt := 0.0;
        for I := J to Ub1 do
          begin
            Sum := A[I,J];
            for K := Lb1 to Pred(J) do
              Sum := Sum - A[I,K] * A[K,J];
            A[I,J] := Sum;
            T := V[I] * Abs(Sum);
            if T > Pvt then
              begin
                Pvt := T;
                Imax := I;
              end;
          end;
        if J <> Imax then
          begin
            for K := Lb1 to Ub1 do
              FSwap(A[Imax,K], A[J,K]);
            V[Imax] := V[J];
          end;
        Index^[J] := Imax;
        if A[J,J] = 0.0 then
          A[J,J] := MachEp;
        if J <> Ub1 then
          begin
            T := 1.0 / A[J,J];
            for I := Succ(J) to Ub1 do
              A[I,J] := A[I,J] * T;
          end;
      end;

    SetErrCode(MatOk);
  end;

procedure LU_Solve(var A : Matrix; var B, X : Vector);
  var
    I, Ip, J, K : Integer;
    Sum         : Float;
  begin
    if (Lb1 <> Lb2) or (Lb <> Lb2) or (Ub1 <> Ub2) or (Ub <> Ub2) then
      begin
        SetErrCode(MatErrDim);
        Exit;
      end;

    for I := Lb to Ub do
      X[I] := B[I];

    K := Pred(Lb);
    for I := Lb to Ub do
      begin
        Ip := Index^[I];
        Sum := X[Ip];
        X[Ip] := X[I];
        if K >= Lb then
          for J := K to Pred(I) do
            Sum := Sum - A[I,J] * X[J]
        else if Sum <> 0.0 then
          K := I;
        X[I] := Sum;
      end;

    for I := Ub downto Lb do
      begin
        Sum := X[I];
        if I < Ub then
          for J := Succ(I) to Ub do
            Sum := Sum - A[I,J] * X[J];
        X[I] := Sum / A[I,I];
      end;

    SetErrCode(MatOk);
  end;

end.
