{ ******************************************************************
  Multiple linear regression (Gauss-Jordan method)
  ****************************************************************** }

unit umulfit;

interface

uses
  utypes, ulineq;

{$i regdef.inc}

procedure MulFit(var X    : Matrix;
                 var Y    : Vector;
                 Nvar     : Integer;
                 ConsTerm : Boolean;
                 var B    : BVector;
                 var V    : VMatrix);
{ ------------------------------------------------------------------
  Multiple linear regression: Y = B(0) + B(1) * X + B(2) * X2 + ...
  ------------------------------------------------------------------
  Input parameters:  X        = matrix of independent variables
                     Y        = vector of dependent variable
                     Nvar     = number of independent variables
                     ConsTerm = presence of constant term B(0)
  Output parameters: B        = regression parameters
                     V        = inverse matrix
  ------------------------------------------------------------------ }

procedure WMulFit(var X    : Matrix;
                  var Y, S : Vector;
                  Nvar     : Integer;
                  ConsTerm : Boolean;
                  var B    : BVector;
                  var V    : VMatrix);
{ ----------------------------------------------------------------------
  Weighted multiple linear regression
  ----------------------------------------------------------------------
  S = standard deviations of observations
  Other parameters as in MulFit
  ---------------------------------------------------------------------- }

implementation

function CheckBounds(var X   : Matrix;
                     var Y   : Vector;
                     Nvar, L : Integer;
                     var B   : BVector;
                     var V   : VMatrix) : Boolean;
var
  Ok : Boolean;
begin
  Ok := (Ub - Lb >= Nvar) and
        (Lb1 = Lb) and (Ub1 = Ub) and
        (Lb2  = 1) and (Ub2  = Nvar) and
        (LbB  = L) and (UbB  = Nvar) and
        (LbV1 = L) and (UbV1 = Nvar) and
        (LbV2 = L) and (UbV2 = Nvar);

  if Ok then SetErrCode(MatOk) else SetErrCode(MatErrDim);

  CheckBounds := Ok;
end;

procedure MulFit(var X    : Matrix;
                 var Y    : Vector;
                 Nvar     : Integer;
                 ConsTerm : Boolean;
                 var B    : BVector;
                 var V    : VMatrix);

var
  L       : Integer;  { Index of first param. (0 if cst term, else 1) }
  I, J, K : Integer;  { Loop variables }
  Det     : Float;    { Determinant }

begin
  if ConsTerm then L := 0 else L := 1;
  if not CheckBounds(X, Y, Nvar, L, B, V) then Exit;

  { Initialize }
  for I := L to Nvar do
    begin
      for J := L to Nvar do
        V[I,J] := 0.0;
      B[I] := 0.0;
    end;

  { If constant term, set line 0 and column 0 of matrix V }
  if ConsTerm then
    begin
      V[0,0] := Ub - Lb + 1;
      for K := Lb to Ub do
        begin
          for J := 1 to Nvar do
            V[0,J] := V[0,J] + X[K,J];
          B[0] := B[0] + Y[K];
        end;
      for J := 1 to Nvar do
        V[J,0] := V[0,J];
    end;

  { Set other elements of V }
  for K := Lb to Ub do
    for I := 1 to Nvar do
      begin
        for J := I to Nvar do
          V[I,J] := V[I,J] + X[K,I] * X[K,J];
        B[I] := B[I] + X[K,I] * Y[K];
      end;

  { Fill in symmetric matrix }
  for I := 2 to Nvar do
    for J := 1 to Pred(I) do
      V[I,J] := V[J,I];

  { Solve normal equations }
  LinEq(V, B, Det);
end;

procedure WMulFit(var X    : Matrix;
                  var Y, S : Vector;
                  Nvar     : Integer;
                  ConsTerm : Boolean;
                  var B    : BVector;
                  var V    : VMatrix);
var
  L       : Integer;  { Index of first param. (0 if cst term, else 1) }
  I, J, K : Integer;  { Loop variables }
  WX      : Float;    { W * X }
  Det     : Float;    { Determinant }

  W : array[Lb..Ub] of Float;  { Weights }

begin
  if ConsTerm then L := 0 else L := 1;
  if not CheckBounds(X, Y, Nvar, L, B, V) then Exit;

  for K := Lb to Ub do
    if S[K] <= 0.0 then
      begin
        SetErrCode(MatSing);
        Exit;
      end
    else
      W[K] := 1.0 / Sqr(S[K]);

  { Initialize }
  for I := L to Nvar do
    begin
      for J := L to Nvar do
        V[I,J] := 0.0;
      B[I] := 0.0;
    end;

  { If constant term, set line 0 and column 0 of matrix V }
  if ConsTerm then
    begin
      for K := Lb to Ub do
        begin
          V[0,0] := V[0,0] + W[K];
          for J := 1 to Nvar do
            V[0,J] := V[0,J] + W[K] * X[K,J];
          B[0] := B[0] + W[K] * Y[K];
        end;
      for J := 1 to Nvar do
        V[J,0] := V[0,J];
    end;

  { Set other elements of V }
  for K := Lb to Ub do
    for I := 1 to Nvar do
      begin
        WX := W[K] * X[K,I];
        for J := I to Nvar do
          V[I,J] := V[I,J] + WX * X[K,J];
        B[I] := B[I] + WX * Y[K];
      end;

  { Fill in symmetric matrix }
  for I := 2 to Nvar do
    for J := 1 to Pred(I) do
      V[I,J] := V[J,I];

  { Solve normal equations }
  LinEq(V, B, Det);
end;

end.
