{ ******************************************************************
  Nonlinear regression
  ****************************************************************** }

unit unlfit;

interface

uses
  utypes, ugausjor, umarq, ubfgs, usimplex,
  usimann, ugenalg, umcmc, ustrings;

{$i regdef.inc}

procedure SetOptAlgo(Algo : TOptAlgo);
{ ------------------------------------------------------------------
  Sets the optimization algorithm according to Algo, which must be
  NL_MARQ, NL_SIMP, NL_BFGS, NL_SA, NL_GA. Default is NL_MARQ
  ------------------------------------------------------------------ }

function GetOptAlgo : TOptAlgo;
{ ------------------------------------------------------------------
  Returns the optimization algorithm
  ------------------------------------------------------------------ }

procedure SetParamBounds(I : Byte; ParamMin, ParamMax : Float);
{ ------------------------------------------------------------------
  Sets the bounds on the I-th regression parameter
  ------------------------------------------------------------------ }
   
procedure GetParamBounds(I : Byte; var ParamMin, ParamMax : Float);  
{ ------------------------------------------------------------------
  Returns the bounds on the I-th regression parameter
  ------------------------------------------------------------------ }

function NullParam(var B : BVector) : Boolean;
{ ------------------------------------------------------------------
  Checks if a regression parameter is equal to zero
  ------------------------------------------------------------------ }

procedure NLFit(RegFunc   : TRegFunc;
                DerivProc : TDerivProc;
                var X, Y  : Vector;
                MaxIter   : Integer;
                Tol       : Float;
                var B     : BVector;
                var V     : VMatrix);
{ ------------------------------------------------------------------
  Unweighted nonlinear regression
  ------------------------------------------------------------------
  Input parameters:  RegFunc   = regression function
                     DerivProc = procedure to compute derivatives
                     X, Y      = point coordinates
                     MaxIter   = max. number of iterations
                     Tol       = tolerance on parameters
                     B         = initial parameter values
  Output parameters: B         = fitted regression parameters
                     V         = inverse matrix
  ------------------------------------------------------------------ }

procedure WNLFit(RegFunc     : TRegFunc;
                 DerivProc   : TDerivProc;
                 var X, Y, S : Vector;
                 MaxIter     : Integer;
                 Tol         : Float;
                 var B       : BVector;
                 var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted nonlinear regression
  ------------------------------------------------------------------
  S = standard deviations of observations
  Other parameters as in NLFit
  ------------------------------------------------------------------ }

procedure SetMCFile(FileName : String);
{ ------------------------------------------------------------------
  Set file for saving MCMC simulations
  ------------------------------------------------------------------ }

procedure SimFit(RegFunc  : TRegFunc;
                 var X, Y : Vector;
                 var B    : BVector;
                 var V    : VMatrix);
{ ------------------------------------------------------------------
  Simulation of unweighted nonlinear regression by MCMC
  ------------------------------------------------------------------ }

procedure WSimFit(RegFunc     : TRegFunc;
                  var X, Y, S : Vector;
                  var B       : BVector;
                  var V       : VMatrix);
{ ------------------------------------------------------------------
  Simulation of weighted nonlinear regression by MCMC
  ------------------------------------------------------------------ }


implementation

const
  MAX_PARAM = 20;       { Max number of parameters }
  MAX_BOUND = 1.0E+6;   { Default parameter bound }
  MAX_FUNC  = 1.0E+30;  { Max. value for objective function
                          (used to prevent overflow) }
var
  OptAlgo : TOptAlgo = NL_MARQ;     { Optimization algorithm }
  MCFile  : String   = 'mcmc.txt';  { File for saving MCMC simulations }

var
  B_min, B_max : array[0..MAX_PARAM] of Float;  { Parameter bounds }

procedure SetOptAlgo(Algo : TOptAlgo);
begin
  OptAlgo := Algo;
end;

function GetOptAlgo : TOptAlgo;
begin
  GetOptAlgo := OptAlgo;
end;

procedure SetParamBounds(I : Byte; ParamMin, ParamMax : Float);
begin
  if (I > MAX_PARAM) or (ParamMin >= ParamMax) then Exit;

  B_min[I] := ParamMin;
  B_max[I] := ParamMax;
end;

procedure GetParamBounds(I : Byte; var ParamMin, ParamMax : Float);
begin
  if I > MAX_PARAM then Exit;

  ParamMin := B_min[I];
  ParamMax := B_max[I];
end;

function NullParam(var B : BVector) : Boolean;
var
  NP : Boolean;
  I  : Integer;
begin
  I := LbB;
  repeat
    NP := (B[I] = 0.0);
    Inc(I);
  until NP or (I > UbB);
  NullParam := NP;
end;

procedure SetGlobalVar(Mode     : TRegMode;
                       var S, W : Vector;
                       var B    : BVector;
                       var V    : VMatrix);

{ Checks the data and sets the global variables }

var
  I : Integer;

begin
  if (Ub - Lb <= UbB - LbB) or (LbB < 0) or (UbB > MAX_PARAM) or
     (LbV1 <> LbB) or (UbV1 <> UbB) or (LbV2 <> LbB) or (UbV2 <> UbB) then
    begin
      SetErrCode(MatErrDim);
      Exit;
    end;

  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    begin
      SetErrCode(NLNullPar);
      Exit;
    end;

  if Mode = WLS then
    for I := Lb to Ub do
      if S[I] <= 0.0 then
        begin
          SetErrCode(MatSing);
          Exit;
        end
      else
        W[I] := 1.0 / Sqr(S[I]);

  for I := LbB to UbB do
    if B_min[I] >= B_max[I] then
      begin
        B_min[I] := - MAX_BOUND;
        B_max[I] :=   MAX_BOUND;
      end;

  SetErrCode(MatOk);
end;

function OutOfBounds(var B : BVector) : Boolean;
{ Check if the parameters are inside the bounds }
var
  I   : Integer;
  OoB : Boolean;
begin
  I := LbB;
  repeat
    OoB := (B[I] < B_min[I]) or (B[I] > B_max[I]);
    Inc(I);
  until OoB or (I > UbB);
  OutOfBounds := OoB;
end;

procedure GenNLFit(Mode        : TRegMode;
                   RegFunc     : TRegFunc;
                   DerivProc   : TDerivProc;
                   var X, Y, S : Vector;
                   MaxIter     : Integer;
                   Tol         : Float;
                   var B       : BVector;
                   var V       : VMatrix);

{ --------------------------------------------------------------------
  General nonlinear regression routine
  -------------------------------------------------------------------- }

var
  F_min   : Float;    { Value of objective function at minimum }
  Det     : Float;    { Determinant of Hessian matrix }
  I, J, K : Integer;  { Loop variables }

  Ycalc, R, W, WR : array[Lb..Ub] of Float;  { Y calc., residuals, weights }

  D : array[LbB..UbB] of Float;            { Derivatives }
  G : array[LbB..UbB] of Float;            { Gradient }
  H : array[LbB..UbB, LbB..UbB] of Float;  { Hessian }

  function ObjFunc(var B : BVector) : Float;
  var
    Sum : Float;
  begin
    if OutOfBounds(B) then
      begin
        ObjFunc := MAX_FUNC;
        Exit;
      end;

    Sum := 0.0;
    K := Lb;

    repeat
      Ycalc[K] := RegFunc(X[K], B);
      R[K] := Y[K] - Ycalc[K];
      if Mode = WLS then WR[K] := W[K] * R[K];

      case Mode of
        OLS : Sum := Sum + Sqr(R[K]);
        WLS : Sum := Sum + WR[K] * R[K];
      end;

      Inc(K);
    until (K > Ub) or (Sum > MAX_FUNC);

    if Sum > MAX_FUNC then Sum := MAX_FUNC;

    ObjFunc := Sum;
  end;

  procedure Gradient(var B, G : BVector);
  begin
    { Initialization }
    for I := LbB to UbB do
      G[I] := 0.0;

    { Compute Gradient }
    for K := Lb to Ub do
      begin
        DerivProc(X[K], Ycalc[K], B, D);

        case Mode of
          OLS : for I := LbB to UbB do G[I] := G[I] - D[I] * R[K];
          WLS : for I := LbB to UbB do G[I] := G[I] - D[I] * WR[K];
        end;
      end;

    for I := LbB to UbB do
      G[I] := 2.0 * G[I];
  end;

  procedure HessGrad(var B, G : BVector; var H : VMatrix);
  { Gradient and Hessian for unweighted nonlinear regression }
  var
    WD : Float;
  begin
    { Initializations }
    for I := LbB to UbB do
      begin
        G[I] := 0.0;
        for J := I to UbB do
          H[I,J] := 0.0;
      end;

    { Compute Gradient & Hessian (both scaled by 1/2) }
    for K := Lb to Ub do
      begin
        DerivProc(X[K], Ycalc[K], B, D);

        case Mode of
          OLS : for I := LbB to UbB do
                  begin
                    G[I] := G[I] - D[I] * R[K];
                    for J := I to UbB do
                      H[I,J] := H[I,J] + D[I] * D[J];
                  end;
          WLS : for I := LbB to UbB do
                  begin
                    G[I] := G[I] - D[I] * WR[K];
                    WD := W[K] * D[I];
                    for J := I to UbB do
                      H[I,J] := H[I,J] + WD * D[J];
                  end;
        end;
      end;

    { Fill in symmetric matrix }
    for I := Succ(LbB) to UbB do
      for J := LbB to Pred(I) do
        H[I,J] := H[J,I];
  end;

begin
  SetGlobalVar(Mode, S, W, B, V);
  if MathErr <> MatOk then Exit;

  case OptAlgo of
    NL_MARQ : Marquardt(ObjFunc, HessGrad, B, MaxIter, Tol, F_min, G, V, Det);
    NL_BFGS : BFGS(ObjFunc, Gradient, B, MaxIter, Tol, F_min, G, V);
    NL_SIMP : Simplex(ObjFunc, B, MaxIter, Tol, F_min);
    NL_SA   : SimAnn(ObjFunc, B, B_min, B_max, F_min);
    NL_GA   : GenAlg(ObjFunc, B, B_min, B_max, F_min);
  end;

  if (OptAlgo <> NL_MARQ) and (MathErr = MatOk) then
    begin
      { Compute the Hessian matrix and its inverse }
      HessGrad(B, G, V);
      GaussJordan(V, Det);
    end;
end;

procedure NLFit(RegFunc   : TRegFunc;
                DerivProc : TDerivProc;
                var X, Y  : Vector;
                MaxIter   : Integer;
                Tol       : Float;
                var B     : BVector;
                var V     : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenNLFit(OLS, RegFunc, DerivProc, X, Y, S, MaxIter, Tol, B, V);
end;

procedure WNLFit(RegFunc     : TRegFunc;
                 DerivProc   : TDerivProc;
                 var X, Y, S : Vector;
                 MaxIter     : Integer;
                 Tol         : Float;
                 var B       : BVector;
                 var V       : VMatrix);
begin
  GenNLFit(WLS, RegFunc, DerivProc, X, Y, S, MaxIter, Tol, B, V);
end;

procedure SetMCFile(FileName : String);
begin
  MCFile := FileName;
end;

procedure GenSimFit(Mode        : TRegMode;
                    RegFunc     : TRegFunc;
                    var X, Y, S : Vector;
                    var B       : BVector;
                    var V       : VMatrix);

var
  NCycles,
  MaxSim,
  SavedSim : Integer;  { Metropolis-Hastings parameters }
  F_min    : Float;    { Value of objective function at minimum }
  R        : Float;    { Range of parameter values }
  I, J     : Integer;  { Loop variables }
  F        : Text;     { File for storing MCMC simulations }

  W : array[Lb..Ub] of Float;  { Weights }

  function ObjFunc(var B : BVector) : Float;
  var
    Sum, R : Float;
    K      : Integer;
  begin
    if OutOfBounds(B) then
      begin
        ObjFunc := MAX_FUNC;
        Exit;
      end;

    Sum := 0.0;
    K := Lb;

    repeat
      R := Y[K] - RegFunc(X[K], B);

      case Mode of
        OLS : Sum := Sum + Sqr(R);
        WLS : Sum := Sum + W[K] * Sqr(R);
      end;

      Inc(K);
    until (K > Ub) or (Sum > MAX_FUNC);

    if Sum > MAX_FUNC then Sum := MAX_FUNC;

    ObjFunc := Sum;
  end;

begin
  { Initialize parameters }
  for I := LbB to UbB do
    B[I] := B_min[I] + 0.5 * (B_max[I] - B_min[I]);

  SetGlobalVar(Mode, S, W, B, V);
  if MathErr <> MatOk then Exit;

  { Initialize variance-covariance matrix }
  { The parameter range is assumed to cover 6 SD's }
  for I := LbB to UbB do
    begin
      R := B_max[I] - B_min[I];
      for J := LbB to UbB do
        if I = J then
          V[I,J] := Sqr(R / 6.0)
        else
          V[I,J] := 0.0;
    end;

  GetMHParams(NCycles, MaxSim, SavedSim);

  var B_opt : array[LbB..UbB] of Float;   { Parameter values at minimum }
  var Xmat  : array[1..SavedSim, LbB..UbB] of Float;  { Simulated param. }

  Hastings(ObjFunc, 2.0, B, V, Xmat, B_opt, F_min);

  { Save simulations }
  if MathErr = MatOk then
    begin
      Assign(F, MCFile);
      Rewrite(F);
      for I := 1 to SavedSim do
        begin
          Write(F, IntStr(I));
          for J := LbB to UbB do
            Write(F, FloatStr(Xmat[I,J]));
          Writeln(F);
        end;
      Close(F);
    end;
end;

procedure SimFit(RegFunc  : TRegFunc;
                 var X, Y : Vector;
                 var B    : BVector;
                 var V    : VMatrix);

var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenSimFit(OLS, RegFunc, X, Y, S, B, V);
end;

procedure WSimFit(RegFunc     : TRegFunc;
                  var X, Y, S : Vector;
                  var B       : BVector;
                  var V       : VMatrix);
begin
  GenSimFit(WLS, RegFunc, X, Y, S, B, V);
end;

end.
