{ ******************************************************************
  Principal component analysis
  ****************************************************************** }

unit upca;

interface

uses
  utypes, ueigsym;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define Matrix array[Lb1..Ub1 : Integer; Lb2..Ub2 : Integer] of Float}

{$define VMatrix array[LbV1..UbV1 : Integer; LbV2..UbV2 : Integer] of Float}

{$define ZMatrix array[LbZ1..UbZ1 : Integer; LbZ2..UbZ2 : Integer] of Float}

procedure VecMean(var X : Matrix; var M : Vector);
{ ------------------------------------------------------------------
  Computes the mean vector (M) from matrix X
  ------------------------------------------------------------------
  Input  : X[Lb1..Ub1, Lb2..Ub2] = matrix of variables (columns)
  Output : M[Lb2..Ub2]           = mean vector
  ------------------------------------------------------------------ }

procedure VecSD(var X : Matrix; var M, S : Vector);
{ ------------------------------------------------------------------
  Computes the vector of standard deviations (S) from matrix X
  ------------------------------------------------------------------
  Input  : X, M
  Output : S[Lb2..Ub2]
  ------------------------------------------------------------------ }

procedure MatVarCov(var X : Matrix; var M : Vector; var V : VMatrix);
{ ------------------------------------------------------------------
  Computes the variance-covariance matrix (V) from matrix X

  Input  : X, M
  Output : V[Lb2..Ub2, Lb2..Ub2]
  ------------------------------------------------------------------ }

procedure MatCorrel(var V, R : Matrix);
{ ------------------------------------------------------------------
  Computes the correlation matrix (R) from the variance-covariance
  matrix (V)

  Input  : V
  Output : R[Lb2..Ub2, Lb2..Ub2]
  ------------------------------------------------------------------ }

procedure PCA(var R : Matrix; var Lambda : Vector; var C, Rc : VMatrix);
{ ------------------------------------------------------------------
  Performs a principal component analysis of the correlation matrix R
  ------------------------------------------------------------------
  Input  : R[Lb2..Ub2, Lb2..Ub2]  = Correlation matrix
  Output : Lambda[Lb2..Ub2]       = Eigenvalues of the correl. matrix
                                    (in descending order)
           C[Lb2..Ub2, Lb2..Ub2]  = Eigenvectors of the correl. matrix
                                    (stored as columns)
           Rc[Lb2..Ub2, Lb2..Ub2] = Correlations between princ. factors
                                    and variables (Rc[I,J] is the
                                    correlation coefficient between
                                    variable I and factor J)
  ------------------------------------------------------------------
  NB : This procedure destroys the original matrix R
  ------------------------------------------------------------------ }

procedure ScaleVar(var X : Matrix; var M, S : Vector; var Z : ZMatrix);
{ ------------------------------------------------------------------
  Scales a set of variables by subtracting means and dividing by SD's
  ------------------------------------------------------------------
  Input  : X, M, S
  Output : Z[Lb..Ub, Lb2..Ub2] = matrix of scaled variables
  ------------------------------------------------------------------ }

procedure PrinFac(var Z : Matrix; var C : VMatrix; var F : ZMatrix);
{ ------------------------------------------------------------------
  Computes principal factors
  ------------------------------------------------------------------
  Input  : Z[Lb..Ub, Lb2..Ub2]   = matrix of scaled variables
           C[Lb2..Ub2, Lb2..Ub2] = matrix of eigenvectors from PCA
  Output : F[Lb..Ub, Lb2..Ub2]   = matrix of principal factors
  ------------------------------------------------------------------ }

implementation

procedure VecMean(var X : Matrix; var M : Vector);
var
  I, K, Nobs : Integer;
  Sum        : Float;
begin
  if (Lb <> Lb2) or (Ub <> Ub2) then
    begin
      SetErrCode(MatErrDim);
      Exit;
    end;

  Nobs := Ub1 - Lb1 + 1;

  for I := Lb2 to Ub2 do
    begin
      Sum := 0.0;
      for K := Lb1 to Ub1 do
        Sum := Sum + X[K,I];
      M[I] := Sum / Nobs;
    end;
end;

procedure VecSD(var X : Matrix; var M, S : Vector);
var
  I, K, Nobs : Integer;
  Sum        : Float;
begin
  if (Lb <> Lb2) or (Ub <> Ub2) then
    begin
      SetErrCode(MatErrDim);
      Exit;
    end;

  Nobs := Ub1 - Lb1 + 1;

  for I := Lb2 to Ub2 do
    begin
      Sum := 0.0;
      for K := Lb1 to Ub1 do
        Sum := Sum + Sqr(X[K,I] - M[I]);
      S[I] := Sqrt(Sum / Nobs);
    end;
end;

procedure MatVarCov(var X : Matrix; var M : Vector; var V : VMatrix);
var
  I, J, K, Nobs : Integer;
  Sum           : Float;
begin
  if (Lb <> Lb2) or (Lb <> LbV1) or (Lb <> LbV2) or
     (Ub <> Ub2) or (Ub <> UbV1) or (Ub <> UbV2) then
     begin
       SetErrCode(MatErrDim);
       Exit;
     end;

  Nobs := Ub1 - Lb1 + 1;

  for I := Lb2 to Ub2 do
    for J := I to Ub2 do
      begin
        Sum := 0.0;
        for K := Lb1 to Ub1 do
          Sum := Sum + (X[K,I] - M[I]) * (X[K,J] - M[J]);
        V[I,J] := Sum / Nobs;
        V[J,I] := V[I,J];
      end;
end;

procedure MatCorrel(var V, R : Matrix);
var
  I, J : Integer;
  P    : Float;
begin
  if (Lb1 <> Lb2) or (Ub1 <> Ub2) then
     begin
       SetErrCode(MatErrDim);
       Exit;
     end;

  for I := Lb2 to Ub2 do
    begin
      R[I,I] := 1.0;
      for J := Succ(I) to Ub2 do
        begin
          P := V[I,I] * V[J,J];
          if P > 0.0 then
            R[I,J] := V[I,J] / Sqrt(P)
          else
            R[I,J] := 0.0;
          R[J,I] := R[I,J];
        end;
    end;
end;

procedure PCA(var R : Matrix; var Lambda : Vector; var C, Rc : VMatrix);
var
  I, J : Integer;
  Rac  : Float;
begin
  if (Lb <> Lb1) or (Lb <> Lb2) or (Lb <> LbV1) or (Lb <> LbV2) or
     (Ub <> Ub1) or (Ub <> Ub2) or (Ub <> UbV1) or (Ub <> UbV2) then
     begin
       SetErrCode(MatErrDim);
       Exit;
     end;

  { Compute eigenvalues and eigenvectors of correlation matrix }
  EigenSym(R, Lambda, C);

  if MathErr <> MatOk then Exit;

  { Compute correlations between principal factors and reduced variables }
  for J := Lb2 to Ub2 do
    begin
      Rac := Sqrt(Lambda[J]);
      for I := Lb2 to Ub2 do
        Rc[I,J] := C[I,J] * Rac;
    end;
end;

procedure ScaleVar(var X : Matrix; var M, S : Vector; var Z : ZMatrix);
var
  I, J : Integer;
begin
  if (Lb <> Lb2) or (Lb <> LbZ2) or
     (Ub <> Ub2) or (Ub <> UbZ2) or
     (Lb1 <> LbZ1) or (Ub1 <> UbZ1) then
     begin
       SetErrCode(MatErrDim);
       Exit;
     end;

  for I := Lb1 to Ub1 do
    for J := Lb2 to Ub2 do
      Z[I,J] := (X[I,J] - M[J]) / S[J];
end;

procedure PrinFac(var Z : Matrix; var C : VMatrix; var F : ZMatrix);
var
  I, J, K : Integer;
begin
  if (Lb2 <> LbV1) or (Lb2 <> LbV2) or (Lb2 <> LbZ2) or
     (Ub2 <> UbV1) or (Ub2 <> UbV2) or (Ub2 <> UbZ2) or
     (Lb1 <> LbZ1) or (Ub1 <> UbZ1) then
     begin
       SetErrCode(MatErrDim);
       Exit;
     end;

  for I := Lb1 to Ub1 do
    for J := Lb2 to Ub2 do
      begin
        F[I,J] := 0.0;
        for K := Lb2 to Ub2 do
          F[I,J] := F[I,J] + Z[I,K] * C[K,J];
      end;
end;

end.
