{ ******************************************************************
  This unit fits the acid/base titration function :

                                    B - A
                       y = A + ----------------
                               1 + 10^(pKa - x)

  where x   is pH
        y   is some property (e.g. absorbance) which depends on the
            ratio of the acidic and basic forms of the compound
        A   is the property for the pure acidic form
        B   is the property for the pure basic form
        pKa is the acidity constant
  ****************************************************************** }

unit upkfit;

interface

uses
  utypes, umath, ulinfit, unlfit;

{$i regdef.inc}

procedure PKFit(var X, Y : Vector;
                MaxIter  : Integer;
                Tol      : Float;
                var B    : BVector;
                var V    : VMatrix);

procedure WPKFit(var X, Y, S : Vector;
                 MaxIter     : Integer;
                 Tol         : Float;
                 var B       : BVector;
                 var V       : VMatrix);

function PKFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Computes the regression function at point X
  B is the vector of parameters, such that :
  B[0] = A     B[1] = B     B[2] = pKa
  ------------------------------------------------------------------ }

implementation

function PKFit_Func(X : Float; var B : BVector) : Float;
begin
  PKFit_Func := B[0] + (B[1] - B[0]) / (1.0 + Exp10(B[2] - X));
end;

procedure PKFit_Deriv(X, Y : Float; var B, D : BVector);
{ ------------------------------------------------------------------
  Computes the derivatives of the regression function at point X
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter.
  ------------------------------------------------------------------ }
var
  Q, R : Float;
begin
  Q := Exp10(B[2] - X);  { 10^(pKa - x) }
  R := 1.0 / (1.0 + Q);  { 1/[1 + 10^(pKa - x)] }

  D[0] := 1.0 - R;       { dy/dA = 1 - 1/[1 + 10^(pKa - x)] }
  D[1] := R;             { dy/dB = 1/[1 + 10^(pKa - x)] }

  { dy/dpKa = (A-B).10^(pKa - x).Ln(10) / [1 + 10^(pKa - x)]^2 }
  D[2] := (B[0] - B[1]) * Q * Ln10 * Sqr(R);
end;

procedure ApproxFit(Mode : TRegMode; var X, Y, S : Vector;
                     var B : BVector);
{ ------------------------------------------------------------------
  Approximate fit of the acid-base titration curve by linear
  regression:
  Ln[(B - A)/(y - A) - 1] = (pKa - x) * Ln(10)
  ------------------------------------------------------------------
  Input :  Mode = OLS for unweighted regression, WLS for weighted
           X, Y = point coordinates
           S    = standard deviations
  Output : B    = estimated regression parameters
  ------------------------------------------------------------------ }
var
  P    : Integer;  { Number of points for linear regression }
  K    : Integer;  { Loop variable }
  Xmin : Float;    { Minimal X coordinate }
  Xmax : Float;    { Maximal X coordinate }
  Imin : Integer;  { Index of Xmin, such that A ~ Y[Imin] }
  Imax : Integer;  { Index of Xmax, such that B ~ Y[Imax] }
  DB   : Float;    { B - A }
  DY   : Float;    { Y - A }
  Z    : Float;    { Transformed Y coordinate }

  A : array[0..1] of Float;        { Linear regression param. }
  V : array[0..1, 0..1] of Float;  { Variance-covariance matrix }

begin
  Xmin := X[Lb]; Imin := Lb;
  Xmax := X[Ub]; Imax := Ub;

  for K := Lb to Ub do
    if X[K] < Xmin then
      begin
        Xmin := X[K];
        Imin := K;
      end
    else if X[K] > Xmax then
      begin
        Xmax := X[K];
        Imax := K;
      end;

  B[0] := Y[Imin];
  B[1] := Y[Imax];

  DB := B[1] - B[0];

  P := 0;
  for K := Lb to Ub do
    if (Y[K] <> B[0]) and (DB / (Y[K] - B[0]) > 1.0) then
      Inc(P);
      
  if P < 4 then
    begin  
      SetErrCode(MatSing);
      Exit;
    end;    
      
  var XX, YY, SS : array[1..P] of Float;
  
  P := 0;
  for K := Lb to Ub do
    if (Y[K] <> B[0]) and (DB / (Y[K] - B[0]) > 1.0) then
      begin
        Inc(P);
        DY := Y[K] - B[0];
        Z := DB / DY - 1.0;
        XX[P] := X[K];
        YY[P] := Ln(Z);
        SS[P] := Abs(DB / (Z * Sqr(DY)));
        if Mode = WLS then SS[P] := SS[P] * S[K];
      end;

  WLinFit(XX, YY, SS, A, V);

  if MathErr = MatOk then
    B[2] := A[0] * InvLn10;
end;

procedure GenPKFit(Mode        : TRegMode;
                   var X, Y, S : Vector;
                   MaxIter     : Integer;
                   Tol         : Float;
                   var B       : BVector;
                   var V       : VMatrix);
begin
  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    ApproxFit(Mode, X, Y, S, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(PKFit_Func, PKFit_Deriv, X, Y,
                MaxIter, Tol, B, V);
    WLS : WNLFit(PKFit_Func, PKFit_Deriv, X, Y, S,
                 MaxIter, Tol, B, V);
  end;
end;

procedure PKFit(var X, Y : Vector;
                MaxIter  : Integer;
                Tol      : Float;
                var B    : BVector;
                var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenPKFit(OLS, X, Y, S, MaxIter, Tol, B, V);
end;

procedure WPKFit(var X, Y, S : Vector;
                 MaxIter     : Integer;
                 Tol         : Float;
                 var B       : BVector;
                 var V       : VMatrix);
begin
  GenPKFit(WLS, X, Y, S, MaxIter, Tol, B, V);
end;

end.
