{ ******************************************************************
  Utility functions to handle roots of polynomials
  ****************************************************************** }

unit upolutil;

interface

uses
  utypes, uswap;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define CompVector array[LbC..UbC : Integer] of Complex}

function SetRealRoots(    Deg : Integer;
                      var Z   : CompVector;
                          Tol : Float) : Integer;
{ ------------------------------------------------------------------
  Set the imaginary part of a root to zero if it is less than a
  fraction Tol of its real part. This root is therefore considered
  real. The function returns the total number of real roots.
  ------------------------------------------------------------------ }

procedure SortRoots(Deg : Integer; var Z : CompVector);
{ ------------------------------------------------------------------
  Sort roots so that:

  (1) The Nr real roots are stored in elements [1..Nr] of vector Z,
      in increasing order.

  (2) The complex roots are stored in elements [(Nr + 1)..Deg] of
      vector Z and are unordered.
  ------------------------------------------------------------------ }

implementation

function SetRealRoots(    Deg : Integer;
                      var Z   : CompVector;
                          Tol : Float) : Integer;
var
  I, N : Integer;
  X, Y : Float;
begin
  for I := 1 to Deg do
    begin
      X := Re(Z[I]);
      Y := Im(Z[I]);
      if (Y <> 0.0) and (Abs(Y) < Tol * Abs(X)) then
        Z[I] := Cmplx(X, 0.0);
    end;

  { Count real roots }
  N := 0;
  for I := 1 to Deg do
    if Im(Z[I]) = 0.0 then
      Inc(N);

  SetRealRoots := N;
end;

procedure SortRoots(Deg : Integer; var Z : CompVector);
var
  I, J, K : Integer;
  Nr, Nc  : Integer;
  R       : array[1..Deg] of Float;
  C       : array[1..Deg] of Complex;

  procedure Sort(var X : Vector; N : Integer);
  { Sort vector X (insertion sort) }
  var
    I, J, K : Integer;
    A       : Float;
  begin
    for I := 1 to Pred(N) do
      begin
        K := I;
        A := X[I];
        for J := Succ(I) to N do
          if X[J] < A then
            begin
              K := J;
              A := X[J];
            end;
        FSwap(X[I], X[K]);
      end;
  end;

begin
  { Count real and complex roots }
  Nr := 0; Nc := 0;
  for I := 1 to Deg do
    if Im(Z[I]) = 0.0 then Inc(Nr) else Inc(Nc);

  { Store real roots in R and complex roots in C }
  J := 0; K := 0;
  for I := 1 to Deg do
    if Im(Z[I]) = 0.0 then
      begin
        Inc(J);
        R[J] := Re(Z[I]);
      end
    else
      begin
        Inc(K);
        C[K] := Z[I];
      end;

  { Sort vector R (insertion sort) }
  if Nr > 0 then Sort(R, Nr);

  { Transfer real roots into elements 1..Nr }
  for I := 1 to Nr do
    Z[I] := Cmplx(R[I], 0.0);

  { Transfer complex roots into elements (Nr+1)..Deg }
  for I := 1 to Nc do
    Z[I + Nr] := C[I];
end;

end.
