{ ******************************************************************
  This unit fits a power function :

                                y = A.x^n

  ****************************************************************** }

unit upowfit;

interface

uses
  utypes, umath, ulinfit, unlfit;

{$i regdef.inc}

procedure PowFit(var X, Y : Vector;
                 MaxIter  : Integer;
                 Tol      : Float;
                 var B    : BVector;
                 var V    : VMatrix);
{ ------------------------------------------------------------------
  Unweighted fit of model
  ------------------------------------------------------------------
  Input parameters:  X, Y    = point coordinates
                     MaxIter = max. number of iterations
                     Tol     = tolerance on parameters
  Output parameters: B       = regression parameters
                     V       = inverse matrix
  ------------------------------------------------------------------ }

procedure WPowFit(var X, Y, S : Vector;
                  MaxIter     : Integer;
                  Tol         : Float;
                  var B       : BVector;
                  var V       : VMatrix);
{ ------------------------------------------------------------------
  Weighted fit of model
  ------------------------------------------------------------------
  Additional input parameter:
  S = standard deviations of observations
  ------------------------------------------------------------------ }

function PowFit_Func(X : Float; var B : BVector) : Float;
{ ------------------------------------------------------------------
  Computes the regression function at point X.
  B is the vector of parameters, such that :

  B[0] = A     B[1] = n
  ------------------------------------------------------------------ }

implementation

const
  FirstParam = 0;
  LastParam  = 1;

function PowFit_Func(X : Float; var B : BVector) : Float;
begin
  PowFit_Func := B[0] * (X ** B[1]);
end;

procedure PowFit_Deriv(X, Y : Float; var B, D : BVector);
{ ------------------------------------------------------------------
  Computes the derivatives of the regression function at point (X,Y)
  with respect to the parameters B. The results are returned in D.
  D[I] contains the derivative with respect to the I-th parameter.
  ------------------------------------------------------------------ }
begin
  D[0] := Y / B[0];    { dy/dA = x^n }
  D[1] := Y * Log(X);  { dy/dk = A.x^n.Ln(x) }
end;

procedure ApproxFit(Mode : TRegMode; var X, Y, S : Vector;
                     var B : BVector);
{ ------------------------------------------------------------------
  Approximate fit of a power function by linear regression:
  Ln(y) = Ln(A) + n.Ln(x)
  ------------------------------------------------------------------
  Input :  Mode = OLS for unweighted regression, WLS for weighted
           X, Y = point coordinates
           S    = standard deviations
           N    = number of points
  Output : B    = estimated regression parameters
  -------------------------------------------------------------------- }
var
  P : Integer;  { Number of points for linear regression }
  K : Integer;  { Loop variable }

  A : array[0..1] of Float;        { Linear regression param. }
  V : array[0..1, 0..1] of Float;  { Variance-covariance matrix }

begin
  { Count number of points with X > 0 and Y > 0 }
  P := Pred(Lb);
  for K := Lb to Ub do
    if (X[K] > 0.0) and (Y[K] > 0.0) then
      Inc(P);

  if P < 3 then
    begin
      SetErrCode(MatSing);
      Exit;
    end;

  var X1, Y1, S1 : array[Lb..P] of Float;

  P := Pred(Lb);
  for K := Lb to Ub do
    if (X[K] > 0.0) and (Y[K] > 0.0) then
      begin
        Inc(P);
        X1[P] := Log(X[K]);
        Y1[P] := Log(Y[K]);
        S1[P] := 1.0 / Y[K];
        if Mode = WLS then S1[P] := S1[P] * S[K];
      end;

  WLinFit(X1, Y1, S1, A, V);

  if MathErr = MatOk then
    begin
      B[0] := Expo(A[0]);
      B[1] := A[1];
    end;
end;

procedure GenPowFit(Mode        : TRegMode;
                    var X, Y, S : Vector;
                    MaxIter     : Integer;
                    Tol         : Float;
                    var B       : BVector;
                    var V       : VMatrix);
begin
  if (GetOptAlgo in [NL_MARQ, NL_BFGS, NL_SIMP]) and NullParam(B) then
    ApproxFit(Mode, X, Y, S, B);

  if MaxIter = 0 then Exit;

  case Mode of
    OLS : NLFit(PowFit_Func, PowFit_Deriv, X, Y,
                MaxIter, Tol, B, V);
    WLS : WNLFit(PowFit_Func, PowFit_Deriv, X, Y, S,
                 MaxIter, Tol, B, V);
  end;
end;

procedure PowFit(var X, Y : Vector;
                 MaxIter  : Integer;
                 Tol      : Float;
                 var B    : BVector;
                 var V    : VMatrix);
var
  S : array[Lb..Ub] of Float;  { Dummy variable }
begin
  GenPowFit(OLS, X, Y, S, MaxIter, Tol, B, V);
end;

procedure WPowFit(var X, Y, S : Vector;
                  MaxIter     : Integer;
                  Tol         : Float;
                  var B       : BVector;
                  var V       : VMatrix);
begin
  GenPowFit(WLS, X, Y, S, MaxIter, Tol, B, V);
end;

end.
