{ ******************************************************************
  Quartic equation
  ****************************************************************** }

unit urtpol4;

interface

uses
  utypes, urtpol2, urtpol3;

{$define Vector array[Lb..Ub : Integer] of Float}

{$define CompVector array[LbC..UbC : Integer] of Complex}

function RootPol4(var Coef : Vector; var Z : CompVector) : Integer;
{ ------------------------------------------------------------------
  Solves the quartic equation:
  Coef[0] + Coef[1] * X + Coef[2] * X^2 + Coef[3] * X^3 +
                                             Coef[4] * X^4 = 0
  ------------------------------------------------------------------ }

implementation

function RootPol4(var Coef : Vector; var Z : CompVector) : Integer;
var
  A, AA, B, C, D : Float;
  Q , R , S      : Float;
  K , KK, L, M   : Float;
  I, N1, N2      : Integer;
  Cf             : array[0..3] of Float;
  Z1, Z2         : array[1..2] of Complex;

  function HighestRealRoot(Deg : Integer; var Z : CompVector) : Float;
  { Find the highest real root among the roots of a polynomial }
  var
    I : Integer;
    R : Float;
  begin
    R := - MaxNum;
    for I := 1 to Deg do
      if (Im(Z[I]) = 0.0) and (Re(Z[I]) > R) then
        R := Re(Z[I]);
    HighestRealRoot := R;
  end;

begin
  for I := 1 to 4 do
    Z[I] := Cmplx(0.0, 0.0);

  if Coef[4] = 0 then
    begin
      RootPol4 := RootPol3(Coef, Z);
      Exit;
    end;

  if Coef[0] = 0.0 then
    begin
      { 0 is root. Equation becomes cubic }
      Cf[0] := Coef[1]; Cf[1] := Coef[2]; Cf[2] := Coef[3];

      { Solve cubic equation }
      RootPol4 := RootPol3(Cf, Z) + 1;

      Exit;
    end;

  if Coef[4] = 1.0 then
    begin
      A := Coef[3] * 0.25;
      B := Coef[2];
      C := Coef[1];
      D := Coef[0];
    end
  else
    begin
      A := Coef[3] / Coef[4] * 0.25;
      B := Coef[2] / Coef[4];
      C := Coef[1] / Coef[4];
      D := Coef[0] / Coef[4];
    end;

  AA := A * A;

  Q := B - 6.0 * AA;
  R := C + A * (8.0 * AA - 2.0 * B);
  S := D - A * C + AA * (B - 3.0 * AA);

  { Compute coefficients of cubic equation }
  Cf[3] := 1.0;
  Cf[2] := 0.5 * Q;
  Cf[1] := 0.25 * (Sqr(Cf[2]) - S);

  { Solve cubic equation and set KK = highest real root }
  if (R = 0.0) and (Cf[1] < 0.0) then
    begin
      { Eq. becomes quadratic with 2 real roots }
      Cf[0] := Cf[1]; Cf[1] := Cf[2]; Cf[2] := 1.0;
      N1 := RootPol2(Cf, Z);
      KK := HighestRealRoot(2, Z);
    end
  else
    begin
      Cf[0] := - 0.015625 * Sqr(R);
      N1 := RootPol3(Cf, Z);
      KK := HighestRealRoot(3, Z);
    end;

  K := Sqrt(KK);
  if K = 0.0 then
    R := Sqrt(Sqr(Q) - 4.0 * S)
  else
    begin
      Q := Q + 4.0 * KK;
      R := 0.5 * R / K;
    end;

  L := 0.5 * (Q - R);
  M := 0.5 * (Q + R);

  { Solve quadratic equation: Y^2 + 2KY + L = 0 }
  Cf[0] := L; Cf[1] := 2.0 * K; Cf[2] := 1.0;
  N1 := RootPol2(Cf, Z1);

  { Solve quadratic equation: Z^2 - 2KZ + M = 0 }
  Cf[0] := M; Cf[1] := -Cf[1];
  N2 := RootPol2(Cf, Z2);

  { Compute roots of quartic polynomial }
  Z[1] := Z1[1] - A;
  Z[2] := Z1[2] - A;
  Z[3] := Z2[1] - A;
  Z[4] := Z2[2] - A;

  RootPol4 := N1 + N2;
end;

end.
